@@ -337,13 +337,13 @@ static int
 PyGProps_length(PyGProps *self)
 {
     GObjectClass *class;
-    int n_props;
+    guint n_props;
     
     class = g_type_class_ref(self->gtype);
     g_object_class_list_properties(class, &n_props);
     g_type_class_unref(class);
     
-    return n_props;
+    return (int)n_props;
 }
 
 static PySequenceMethods _PyGProps_as_sequence = {
@@ -558,6 +558,8 @@ pygobject_new_with_interfaces(GType gtype)
     PyObject *modules, *module;
     gchar *type_name, *mod_name, *gtype_name;
 
+    state = pyg_gil_state_ensure();
+
     interfaces = g_type_interfaces (gtype, &n_interfaces);
     bases = PyTuple_New(n_interfaces+1);
     
@@ -566,6 +568,7 @@ pygobject_new_with_interfaces(GType gtype)
     py_parent_type = pygobject_lookup_class(parent_type);
 
     /* We will always put the parent at the first position in bases */
+    Py_INCREF(py_parent_type); /* PyTuple_SetItem steals a reference */
     PyTuple_SetItem(bases, 0, (PyObject*)py_parent_type);
 
     /* And traverse interfaces */
@@ -573,6 +576,7 @@ pygobject_new_with_interfaces(GType gtype)
 	for (i = 0; i < n_interfaces; i++) {
 	    interface_type = interfaces[i];
 	    py_interface_type = pygobject_lookup_class(interface_type);
+            Py_INCREF(py_interface_type); /* PyTuple_SetItem steals a reference */
 	    PyTuple_SetItem(bases, i+1, (PyObject*)py_interface_type);
 	}
 	
@@ -611,16 +615,14 @@ pygobject_new_with_interfaces(GType gtype)
 	type_name = g_strconcat(mod_name, ".", gtype_name, NULL);
     }
 
-    state = pyg_gil_state_ensure();
 
-    type = (PyTypeObject*)PyObject_CallFunction((PyObject*)&PyType_Type, "sOO",
+    type = (PyTypeObject*)PyObject_CallFunction((PyObject*)&PyType_Type, "sNN",
 						type_name, bases, dict);
     g_free(type_name);
 
-    pyg_gil_state_release(state);
-    
     if (type == NULL) {
 	PyErr_Print();
+        pyg_gil_state_release(state);
 	return NULL;
     }
 
@@ -637,6 +639,7 @@ pygobject_new_with_interfaces(GType gtype)
 
     if (PyType_Ready(type) < 0) {
 	g_warning ("couldn't make the type `%s' ready", type->tp_name);
+        pyg_gil_state_release(state);
 	return NULL;
     }
     
@@ -651,6 +654,8 @@ pygobject_new_with_interfaces(GType gtype)
     Py_INCREF(type);
     g_type_set_qdata(gtype, pygobject_class_key, type);
 
+    pyg_gil_state_release(state);
+
     return type;
 }
 
@@ -997,6 +1002,14 @@ pygobject__gobject_init__(PyGObject *self, PyObject *args, PyObject *kwargs)
     return Py_None;
 }
 
+#define CHECK_GOBJECT(self) \
+    if (!G_IS_OBJECT(self->obj)) {                                           \
+	PyErr_Format(PyExc_TypeError,                                        \
+                     "object at 0x%x of type %s is not initialized",	     \
+                     (int)self, self->ob_type->tp_name);                     \
+	return NULL;                                                         \
+    }
+
 static PyObject *
 pygobject_get_property(PyGObject *self, PyObject *args)
 {
@@ -1007,6 +1020,9 @@ pygobject_get_property(PyGObject *self, PyObject *args)
 
     if (!PyArg_ParseTuple(args, "s:GObject.get_property", &param_name))
 	return NULL;
+
+    CHECK_GOBJECT(self);
+    
     pspec = g_object_class_find_property(G_OBJECT_GET_CLASS(self->obj),
 					 param_name);
     if (!pspec) {
@@ -1036,6 +1052,9 @@ pygobject_set_property(PyGObject *self, PyObject *args)
     if (!PyArg_ParseTuple(args, "sO:GObject.set_property", &param_name,
 			  &pvalue))
 	return NULL;
+    
+    CHECK_GOBJECT(self);
+    
     pspec = g_object_class_find_property(G_OBJECT_GET_CLASS(self->obj),
 					 param_name);
     if (!pspec) {
@@ -1056,6 +1075,9 @@ pygobject_freeze_notify(PyGObject *self, PyObject *args)
 {
     if (!PyArg_ParseTuple(args, ":GObject.freeze_notify"))
 	return NULL;
+    
+    CHECK_GOBJECT(self);
+    
     g_object_freeze_notify(self->obj);
     Py_INCREF(Py_None);
     return Py_None;
@@ -1068,6 +1090,9 @@ pygobject_notify(PyGObject *self, PyObject *args)
 
     if (!PyArg_ParseTuple(args, "s:GObject.notify", &property_name))
 	return NULL;
+    
+    CHECK_GOBJECT(self);
+    
     g_object_notify(self->obj, property_name);
     Py_INCREF(Py_None);
     return Py_None;
@@ -1078,6 +1103,9 @@ pygobject_thaw_notify(PyGObject *self, PyObject *args)
 {
     if (!PyArg_ParseTuple(args, ":GObject.thaw_notify"))
 	return NULL;
+    
+    CHECK_GOBJECT(self);
+    
     g_object_thaw_notify(self->obj);
     Py_INCREF(Py_None);
     return Py_None;
@@ -1092,6 +1120,9 @@ pygobject_get_data(PyGObject *self, PyObject *args)
 
     if (!PyArg_ParseTuple(args, "s:GObject.get_data", &key))
 	return NULL;
+    
+    CHECK_GOBJECT(self);
+    
     quark = g_quark_from_string(key);
     data = g_object_get_qdata(self->obj, quark);
     if (!data) data = Py_None;
@@ -1108,6 +1139,9 @@ pygobject_set_data(PyGObject *self, PyObject *args)
 
     if (!PyArg_ParseTuple(args, "sO:GObject.set_data", &key, &data))
 	return NULL;
+    
+    CHECK_GOBJECT(self);
+    
     quark = g_quark_from_string(key);
     Py_INCREF(data);
     g_object_set_qdata_full(self->obj, quark, data, pyg_destroy_notify);
@@ -1140,9 +1174,14 @@ pygobject_connect(PyGObject *self, PyObject *args)
 	PyErr_SetString(PyExc_TypeError, "second argument must be callable");
 	return NULL;
     }
+    
+    CHECK_GOBJECT(self);
+    
     if (!g_signal_parse_name(name, G_OBJECT_TYPE(self->obj),
 			     &sigid, &detail, TRUE)) {
-	PyErr_SetString(PyExc_TypeError, "unknown signal name");
+	PyErr_Format(PyExc_TypeError, "%s: unknown signal name: %s",
+		     PyString_AsString(PyObject_Repr((PyObject*)self)),
+		     name);
 	return NULL;
     }
     extra_args = PySequence_GetSlice(args, 2, len);
@@ -1182,9 +1221,14 @@ pygobject_connect_after(PyGObject *self, PyObject *args)
 	PyErr_SetString(PyExc_TypeError, "second argument must be callable");
 	return NULL;
     }
+    
+    CHECK_GOBJECT(self);
+    
     if (!g_signal_parse_name(name, G_OBJECT_TYPE(self->obj),
 			     &sigid, &detail, TRUE)) {
-	PyErr_SetString(PyExc_TypeError, "unknown signal name");
+	PyErr_Format(PyExc_TypeError, "%s: unknown signal name: %s",
+		     PyString_AsString(PyObject_Repr((PyObject*)self)),
+		     name);
 	return NULL;
     }
     extra_args = PySequence_GetSlice(args, 2, len);
@@ -1224,9 +1268,14 @@ pygobject_connect_object(PyGObject *self, PyObject *args)
 	PyErr_SetString(PyExc_TypeError, "second argument must be callable");
 	return NULL;
     }
+    
+    CHECK_GOBJECT(self);
+    
     if (!g_signal_parse_name(name, G_OBJECT_TYPE(self->obj),
 			     &sigid, &detail, TRUE)) {
-	PyErr_SetString(PyExc_TypeError, "unknown signal name");
+	PyErr_Format(PyExc_TypeError, "%s: unknown signal name: %s",
+		     PyString_AsString(PyObject_Repr((PyObject*)self)),
+		     name);
 	return NULL;
     }
     extra_args = PySequence_GetSlice(args, 3, len);
@@ -1266,9 +1315,14 @@ pygobject_connect_object_after(PyGObject *self, PyObject *args)
 	PyErr_SetString(PyExc_TypeError, "second argument must be callable");
 	return NULL;
     }
+    
+    CHECK_GOBJECT(self);
+    
     if (!g_signal_parse_name(name, G_OBJECT_TYPE(self->obj),
 			     &sigid, &detail, TRUE)) {
-	PyErr_SetString(PyExc_TypeError, "unknown signal name");
+	PyErr_Format(PyExc_TypeError, "%s: unknown signal name: %s",
+		     PyString_AsString(PyObject_Repr((PyObject*)self)),
+		     name);
 	return NULL;
     }
     extra_args = PySequence_GetSlice(args, 3, len);
@@ -1289,6 +1343,9 @@ pygobject_disconnect(PyGObject *self, PyObject *args)
 
     if (!PyArg_ParseTuple(args, "i:GObject.disconnect", &handler_id))
 	return NULL;
+    
+    CHECK_GOBJECT(self);
+    
     g_signal_handler_disconnect(self->obj, handler_id);
     Py_INCREF(Py_None);
     return Py_None;
@@ -1302,6 +1359,9 @@ pygobject_handler_is_connected(PyGObject *self, PyObject *args)
     if (!PyArg_ParseTuple(args, "i:GObject.handler_is_connected", &handler_id))
 	return NULL;
 
+    
+    CHECK_GOBJECT(self);
+    
     return PyBool_FromLong(g_signal_handler_is_connected(self->obj, handler_id));
 }
 
@@ -1312,6 +1372,9 @@ pygobject_handler_block(PyGObject *self, PyObject *args)
 
     if (!PyArg_ParseTuple(args, "i:GObject.handler_block", &handler_id))
 	return NULL;
+    
+    CHECK_GOBJECT(self);
+    
     g_signal_handler_block(self->obj, handler_id);
     Py_INCREF(Py_None);
     return Py_None;
@@ -1350,9 +1413,14 @@ pygobject_emit(PyGObject *self, PyObject *args)
 	return NULL;
     }
     Py_DECREF(first);
+    
+    CHECK_GOBJECT(self);
+    
     if (!g_signal_parse_name(name, G_OBJECT_TYPE(self->obj),
 			     &signal_id, &detail, TRUE)) {
-	PyErr_SetString(PyExc_TypeError, "unknown signal name");
+	PyErr_Format(PyExc_TypeError, "%s: unknown signal name: %s",
+		     PyString_AsString(PyObject_Repr((PyObject*)self)),
+		     name);
 	return NULL;
     }
     g_signal_query(signal_id, &query);
@@ -1420,9 +1488,14 @@ pygobject_stop_emission(PyGObject *self, PyObject *args)
 
     if (!PyArg_ParseTuple(args, "s:GObject.stop_emission", &signal))
 	return NULL;
+    
+    CHECK_GOBJECT(self);
+    
     if (!g_signal_parse_name(signal, G_OBJECT_TYPE(self->obj),
 			     &signal_id, &detail, TRUE)) {
-	PyErr_SetString(PyExc_TypeError, "unknown signal name");
+	PyErr_Format(PyExc_TypeError, "%s: unknown signal name: %s",
+		     PyString_AsString(PyObject_Repr((PyObject*)self)),
+		     signal);
 	return NULL;
     }
     g_signal_stop_emission(self->obj, signal_id, detail);
@@ -1439,7 +1512,9 @@ pygobject_chain_from_overridden(PyGObject *self, PyObject *args)
     const gchar *name;
     GSignalQuery query;
     GValue *params, ret = { 0, };
-
+    
+    CHECK_GOBJECT(self);
+    
     ihint = g_signal_get_invocation_hint(self->obj);
     if (!ihint) {
 	PyErr_SetString(PyExc_TypeError, "could not find signal invocation "
@@ -1607,30 +1682,28 @@ PyTypeObject PyGObject_Type = {
 
 
 static int
-pygobjectmeta_register(PyTypeObject *subtype, PyObject *instance_dict)
+pygobjectmeta_register(PyTypeObject *subtype)
 {
     PyObject *pytype_name;
-    char *type_name = NULL;
+    const char *type_name = NULL;
     int retval = 0;
     
-    if (!instance_dict) {
-        PyErr_Clear();
-	goto out;
-    }
-
     /* Maybe the type doesn't really need to registered? */
-    if (!(PyDict_GetItemString(instance_dict, "__gtype_name__")
-          || PyDict_GetItemString(instance_dict, "__gproperties__")
-          || PyDict_GetItemString(instance_dict, "__gsignals__")))
+    if (subtype->tp_dict == NULL)
+	goto out;
+	
+    if (!(PyDict_GetItemString(subtype->tp_dict, "__gtype_name__")
+          || PyDict_GetItemString(subtype->tp_dict, "__gproperties__")
+          || PyDict_GetItemString(subtype->tp_dict, "__gsignals__")))
 	goto out;
 
     /* If it's already registered, skip registration */
-    if (PyDict_GetItemString(instance_dict, "__gtype__"))
+    if (PyDict_GetItemString(subtype->tp_dict, "__gtype__"))
 	goto out;
 
-    pytype_name = PyDict_GetItemString(instance_dict, "__gtype_name__");
+    pytype_name = PyDict_GetItemString(subtype->tp_dict, "__gtype_name__");
     if (pytype_name)
-	type_name = g_strdup(PyString_AsString(pytype_name));
+	type_name = PyString_AsString(pytype_name);
     retval = pyg_type_register(subtype, type_name);
 
 out:
@@ -1641,13 +1714,10 @@ out:
 static int
 pygobjectmeta_init(PyTypeObject *subtype, PyObject *args, PyObject *kwargs)
 {
-    PyObject *instance_dict;
-    
     if (PyType_Type.tp_init((PyObject *) subtype, args, kwargs))
         return -1;
     
-    instance_dict = PyTuple_GetItem(args, 2);
-    return pygobjectmeta_register(subtype, instance_dict);
+    return pygobjectmeta_register(subtype);
 }
 
 
